import numpy as np
from scipy.spatial.distance import cdist

from pymoo.model.indicator import Indicator
from pymoo.util.non_dominated_rank import NonDominatedRank


class RMetric(Indicator):
    def __init__(self, curr_pop, whole_pop, ref_points, problem, w=None):
        """
        RMetric Constructor
        :param curr_pop: numpy array: Population from algorithm being evaluated
        :param whole_pop: numpy array: Whole population of all algorithms
        :param ref_points: numpy array: list of reference points
        :param problem: problem: problem
        :param w: numpy array: weights for each objective
        """
        Indicator.__init__(self)
        self.curr_pop = curr_pop
        self.whole_pop = whole_pop
        self.ref_points = ref_points
        self.problem = problem
        w_ = np.ones(self.ref_points.shape[1]) if not w else w
        self.w_points = self.ref_points + 2 * w_


    def _filter(self):

        def check_dominance(a, b, n_obj):
            flag1 = False
            flag2 = False
            for i in range(n_obj):
                if a[i] < b[i]:
                    flag1 = True
                else:
                    if a[i] > b[i]:
                        flag2 = True
            if flag1 and not flag2:
                return 1
            elif not flag1 and flag2:
                return -1
            else:
                return 0

        num_objs = np.size(self.curr_pop, axis=1)
        index_array = np.zeros(np.size(self.curr_pop, axis=0))
        for i in range(np.size(self.curr_pop, 0)):
            for j in range(np.size(self.whole_pop, 0)):
                flag = check_dominance(self.curr_pop[i, :], self.whole_pop[j, :], num_objs)
                if flag == -1:
                    index_array[i] = 1
                    break
        final_index = np.logical_not(index_array)
        filtered_pop = self.curr_pop[final_index, :]

        return filtered_pop

    def _filter_fast(self):
        filtered_pop = NonDominatedRank.get_non_dominated(self.whole_pop, self.curr_pop)
        return filtered_pop

    def _preprocess(self, data, ref_point, w_point):

        datasize = np.size(data, 0)

        # Identify representative point
        ref_matrix = np.tile(ref_point, (datasize, 1))
        w_matrix = np.tile(w_point, (datasize, 1))
        # ratio of distance to the ref point over the distance between the w_point and the ref_point
        diff_matrix = (data-ref_matrix)/(w_matrix-ref_matrix)
        agg_value = np.amax(diff_matrix, axis=1)
        idx = np.argmin(agg_value)
        zp = [data[idx, :]]

        return zp,


    def _translate(self, zp, trimmed_data, ref_point, w_point):
        # Solution translation - Matlab reproduction
        # find k
        temp = ((zp[0]-ref_point)/(w_point - ref_point))
        kIdx = np.argmax(temp)

        # find zl
        temp = (zp[0][kIdx] - ref_point[kIdx])/(w_point[kIdx] - ref_point[kIdx])
        zl = ref_point + temp*(w_point - ref_point)

        temp = zl - zp
        shift_direction = np.tile(temp, (trimmed_data.shape[0], 1))
        # new_size = self.curr_pop.shape[0]
        return trimmed_data + shift_direction


    def _trim(self, pop, centeroid, range=0.2):
        """
        Box trimming
        :param pop:
        :param centeroid:
        :param range:
        :return:
        """
        popsize, objDim = pop.shape
        diff_matrix = pop - np.tile(centeroid,(popsize, 1))[0]
        flags = np.sum(abs(diff_matrix) < range/2, axis=1)
        filtered_matrix = pop[np.where(flags == objDim)]
        return filtered_matrix

    def _trim_fast(self, pop, centeroid, range=0.2):
        """
        Euclidean trimming
        :param pop:
        :param centeroid:
        :param range:
        :return:
        """
        centeroid_matrix = cdist(pop, centeroid, metric='euclidean')
        filtered_matrix = pop[np.where(centeroid_matrix < range/2), :][0]
        return filtered_matrix

    def calc(self, hyper_volume=True):
        """
        This method calculates the R-IGD and R-HV based off of the population that was provided
        :return: R-IGD and R-HV
        """
        translated = []
        final_PF = []

        # 1. Prescreen Procedure - NDS Filtering
        pop = self._filter()

        solution = self.problem.calc_pareto_front()
        # solution = calc_PF(1, 10000, 2)

        labels = np.argmin(cdist(pop, self.ref_points), axis=1)

        for i in range(len(self.ref_points)):
            cluster = pop[np.where(labels == i)]
            if len(cluster) != 0:

                # 2. Representative Point Identification
                zp = self._preprocess(cluster, self.ref_points[i], w_point=self.w_points[i])[0]
                # 3. Filtering Procedure - Filter points
                trimmed_data = self._trim(cluster, zp)
                # trimmed_data = self.trim_fast(data, zp, radius)  -- Broken
                # 4. Solution Translation
                pop_t = self._translate(zp, trimmed_data, self.ref_points[i], w_point=self.w_points[i])
                translated.extend(pop_t)

            # 5. R-Metric Computation
            target = self._preprocess(data=solution, ref_point=self.ref_points[i], w_point=self.w_points[i])
            PF = self._trim(solution, target)
            final_PF.extend(PF)

        translated = np.array(translated)

        if np.size(translated) == 0:
            igd = -1
            volume= -1
        else:
            # IGD Computation
            from pymoo.indicators.igd import IGD
            IGD_ = IGD(final_PF)
            igd = IGD_.calc(translated)
            # HV Computation

            nadir_point = np.amax(self.w_points, axis=0)
            front = translated
            dim = self.ref_points[0].shape[0]
            if hyper_volume:
                if dim <3:
                    try:
                        #Python
                        from pymoo.indicators.hv import HyperVolume
                        hv = HyperVolume(nadir_point)
                        volume = hv.compute(front)
                    except TypeError:
                        volume = -1

                else:
                    # cpp

                    from pymoo.cpp.hypervolume.build import hypervolume

                    volume = hypervolume.calculate(dim, len(front), front, nadir_point)
            else:
                volume = np.nan
        return igd, volume


# Matlab code PF calculation method
def calc_PF(problem_id, sample_size=10000, objDim=None):
    # ZDT 1
    if problem_id == 1:
        objDim = 2
        step = 1/(sample_size - 1)
        P = np.zeros((sample_size, objDim))
        f1 = np.arange(0, 1+step, step)[:, None]
        P[:, 0] = f1.T
        P[:, 1] = (np.ones((sample_size, 1)) - np.sqrt(P[:, 0])[:, None])[:,0]
        return P
    # ZDT 2
    if problem_id == 2:
        objDim = 2
        step = 1/(sample_size - 1)
        P = np.zeros((sample_size, objDim))
        f1 = np.arange(0, 1+step, step)[:, None]
        P = np.array([f1, 1 - np.power(f1, 2)]).T
        return P



import numpy as np


def calc_as_fronts_pygmo(pop):

    import pygmo as pg

    if pop.size() == 0:
        return np.array()

    # if there are constraints
    if pop.G.shape[1] > 0:
        constr = np.sum(pop.G[pop.G > 0], axis=1)

        uconstr = np.unique(constr)
        np.sort(uconstr)

        res = []

        for violation in uconstr:
            indices = np.where(constr == violation)[0]

            if len(indices) > 1:
                to_sort = pop.F[indices]
                fronts = pg.fast_non_dominated_sorting(to_sort)[0]
                for front in fronts:
                    mapped = [indices[front[j]] for j in range(len(front))]
                    res.append(np.array(mapped))
            else:
                res.append(indices)

        return res

    else:
        return pg.fast_non_dominated_sorting(pop.F)[0]


class Evaluator:

    def __init__(self, n_eval=10000):
        self.n_eval = n_eval
        self.counter = 0

    def eval(self, problem, x, **kwargs):
        if len(x.shape) == 1:
            self.counter += 1
        else:
            self.counter += x.shape[0]
        return problem.evaluate(x, **kwargs)

    def count_left(self):
        return self.n_eval - self.counter

    def has_next(self):
        return self.counter < self.n_eval



import unittest

import numpy as np

from pymoo.operators.survival.rank_and_crowding import RankAndCrowdingSurvival


class NonDominatedRankTest(unittest.TestCase):
    def test_crowding_distance(self):
        F = np.array([[0.31, 6.10], [0.22, 7.09], [0.79, 3.97], [0.27, 6.93]])
        cd = RankAndCrowdingSurvival.calc_crowding_distance(F, F_min=np.array([0.1, 0]), F_max=np.array([1, 60]))
        self.assertTrue(np.all(np.round(cd, decimals=2) == np.array([0.63, np.inf, np.inf, 0.12])))

    def test_crowding_distance_degenerated(self):
        F = np.array([[0.0, 6.10], [0.0, 7.09], [0.0, 7.09]])
        cd = RankAndCrowdingSurvival.calc_crowding_distance(F)

        print(cd)

if __name__ == '__main__':
    unittest.main()



from abc import abstractmethod
import numpy as np

from pymop.problem import Problem


class Sampling:
    """

    This abstract class represents any sampling strategy that can be used
    to create an initial population or just an initial point

    """

    @abstractmethod
    def sample(self, problem, n_samples, data=None):
        """
        Sample new points according to the problem.


        Parameters
        ----------
        problem: class
            The problem to which points should be sampled. (lower and upper bounds, discrete, binary, ...)

        n_samples: int
            Number of points that should be created

        data: class
            Any additional data that might be necessary. e.g. constants of the algorithm, ...

        Returns
        -------

        """
        pass

    def sample_by_bounds(self, x_min, x_max, n_var, n_samples, data=None):
        class P(Problem):
            def __init__(self) -> None:
                self.n_var = n_var
                self.xl = np.full(n_var, x_min)
                self.xu = np.full(n_var, x_max)

        return self.sample(P(), n_samples, data)




import numpy as np
from pymoo.model.survival import Survival
from pymoo.rand import random
from pymoo.util.non_dominated_rank import NonDominatedRank
from scipy.spatial.distance import cdist


class ReferencePointSurvival(Survival):
    def __init__(self, ref_points, epsilon, weights=None, F_min=None, F_max=None):
        super().__init__()
        self.ref_points = ref_points
        self.epsilon = epsilon
        self.weights = weights
        self.F_min = F_min
        self.F_max = F_max

    def _do(self, pop, n_survive, data, return_only_index=False):

        fronts = NonDominatedRank.calc_as_fronts(pop.F, pop.G)

        # all indices to survive
        survival = []

        for front in fronts:
            if len(survival) + len(front) > n_survive:
                break
            survival.extend(front)

        # filter the front to only relevant entries
        pop.filter(survival + front)
        survival = list(range(0, len(survival)))
        last_front = np.arange(len(survival), pop.size())
        # Indices of last front members that survived
        survived = []

        # if the last front needs to be splitted
        n_remaining = n_survive - len(survival)

        if n_remaining > 0:

            F_min = pop.F.min(axis=0) if self.F_min is None else self.F_min
            F_max = pop.F.max(axis=0) if self.F_max is None else self.F_max

            # dist_matrix = calc_perpendicular_dist_matrix(N, self.ref_dirs)
            dist_matrix = calc_ref_dist_matrix(pop.F, self.ref_points, F_min, F_max, weights=self.weights)
            # point_distance_matrix = cdist(N[last_front, :], N[last_front, :])
            point_distance_matrix = calc_ref_dist_matrix(pop.F[last_front, :], pop.F[last_front, :], F_min, F_max,
                                                         weights=self.weights)
            niche_of_individuals = np.argmin(dist_matrix, axis=1)
            min_dist_matrix = dist_matrix[np.arange(len(dist_matrix)), niche_of_individuals]

            # for each reference direction the niche count
            niche_count = np.zeros(len(self.ref_points))
            for i in niche_of_individuals[survival]:
                niche_count[i] += 1

            # relative index now to dist and the niches
            min_dist_matrix = min_dist_matrix[last_front]
            niche_of_individuals = niche_of_individuals[last_front]

            # boolean array of elements that survive if true
            survival_last_front = np.full(len(last_front), False)

            while n_remaining > 0:

                # all niches where new individuals can be assigned to
                next_niches_list = np.unique(niche_of_individuals[np.logical_not(survival_last_front)])

                # pick a niche with minimum assigned individuals - break tie if necessary
                next_niche_count = niche_count[next_niches_list]
                next_niche = np.where(next_niche_count == next_niche_count.min())[0]
                next_niche = next_niche[random.randint(0, len(next_niche))]
                next_niche = next_niches_list[next_niche]

                # indices of individuals in last front to assign niche to
                next_ind = np.where(niche_of_individuals[np.logical_not(survival_last_front)] == next_niche)[0]
                next_ind = np.where(np.logical_not(survival_last_front))[0][next_ind]

                # Pick the closest point
                next_ind = next_ind[np.argmin(min_dist_matrix[next_ind])]

                # Find surrounding points within trust region
                surrounding_points = np.where(point_distance_matrix[next_ind] < self.epsilon)[0]

                # Clear points in trust region
                survival_last_front[surrounding_points] = True

                # Add selected point to survived population
                survived.append(next_ind)

                if np.all(survival_last_front):
                    survival_last_front = np.full(len(last_front), False)
                    survival_last_front[survived] = True

                niche_count[next_niche] += 1
                n_remaining -= 1

        survival.extend(last_front[survived])
        if return_only_index:
            return survival

        # now truncate the population
        pop.filter(survival)

        return pop


def calc_ref_dist_matrix(F, ref_points, F_min, F_max, weights=None):
    if weights is None:
        weights = np.full(F.shape[1], 1 / F.shape[1])

    r = np.tile(ref_points, (len(F), 1))
    f = np.repeat(F, len(ref_points), axis=0)

    matrix = np.sqrt(np.sum(weights * ((f - r) / (F_max - F_min)) ** 2, axis=1))
    matrix = np.reshape(matrix, (len(F), len(ref_points)))
    return matrix



from pymoo.model.mutation import Mutation
from pymoo.rand import random


class PolynomialMutation(Mutation):
    def __init__(self, eta_mut=20, p_mut=None):
        self.eta_m = eta_mut
        self.p_mut = p_mut

    def _do(self, p, X, Y):

        if self.p_mut is None:
            self.p_mut = 1.0 / p.n_var

        for i in range(X.shape[0]):

            for j in range(X.shape[1]):

                # no mutation for this index
                if random.random() > self.p_mut:
                    Y[i, j] = X[i, j]
                else:
                    y = X[i, j]
                    yl = p.xl[j]
                    yu = p.xu[j]
                    delta1 = (y - yl) / (yu - yl)
                    delta2 = (yu - y) / (yu - yl)
                    r = random.random()
                    mut_pow = 1.0 / (self.eta_m + 1.0)
                    if r <= 0.5:
                        xy = 1.0 - delta1
                        val = 2.0 * r + (1.0 - 2.0 * r) * (pow(xy, (self.eta_m + 1.0)))
                        deltaq = pow(val, mut_pow) - 1.0

                    else:
                        xy = 1.0 - delta2
                        val = 2.0 * (1.0 - r) + 2.0 * (r - 0.5) * (pow(xy, (self.eta_m + 1.0)))
                        deltaq = 1.0 - (pow(val, mut_pow))

                    y = y + deltaq * (yu - yl)
                    if y < yl:
                        y = yl
                    if y > yu:
                        y = yu
                    Y[i, j] = y

        return Y



from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.default_operators import set_default_if_none, set_if_none
from pymoo.operators.survival.fitness_survival import FitnessSurvival


class SingleObjectiveGeneticAlgorithm(GeneticAlgorithm):

    def __init__(self, var_type, **kwargs):
        set_default_if_none(var_type, kwargs)
        set_if_none(kwargs, 'survival', FitnessSurvival())
        super().__init__(**kwargs)



from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.default_operators import set_default_if_none, set_if_none
from pymoo.operators.selection.tournament_selection import TournamentSelection
from pymoo.operators.survival.rank_and_crowding import RankAndCrowdingSurvival
from pymoo.rand import random
from pymoo.util.dominator import Dominator


class NSGAII(GeneticAlgorithm):
    def __init__(self, var_type, **kwargs):
        set_default_if_none(var_type, kwargs)
        set_if_none(kwargs, 'selection', TournamentSelection(f_comp=comp_by_rank_and_crowding))
        set_if_none(kwargs, 'survival', RankAndCrowdingSurvival())
        super().__init__(**kwargs)


def comp_by_rank_and_crowding(pop, indices, data):
    if len(indices) != 2:
        raise ValueError("Only implemented for binary tournament!")

    first = indices[0]
    second = indices[1]

    if data.rank[first] < data.rank[second]:
        return first
    elif data.rank[second] < data.rank[first]:
        return second
    else:
        if data.crowding[first] > data.crowding[second]:
            return first
        elif data.crowding[second] > data.crowding[first]:
            return second
        else:
            return indices[random.randint(0, 2)]


def comp_by_dom_and_crowding(pop, indices, data):
    if len(indices) != 2:
        raise ValueError("Only implemented for binary tournament!")

    first = indices[0]
    second = indices[1]

    rel = Dominator.get_relation(pop.F[first, :], pop.F[second, :])

    if rel == 1:
        return first
    elif rel == -1:
        return second
    else:
        if data.crowding[first] > data.crowding[second]:
            return first
        elif data.crowding[second] > data.crowding[first]:
            return second
        else:
            return indices[random.randint(0, 2)]



import numpy as np

from pymoo.model.algorithm import Algorithm
from pymoo.model.population import Population
from pymoo.util.misc import unique_rows


class GeneticAlgorithm(Algorithm):
    """

    This class represents a basic genetic algorithm that can be extended and modified by
    providing different modules in the constructor. Most importantly the modules for the initial sampling,
    recombination and the survival selection should be provided.
    Different implementations of algorithm can use this class.

    Attributes
    ----------

    pop_size: int
        The population size to be used for the genetic algorithm.

    sampling : class
        The sampling implementation to create the initial population.

    selection : class
        A class to select the parents for the crossover

    crossover : class
        The crossover to be performed on parents.

    mutation : class
        The mutation that will be performed for each child after the crossover

    survival : class
        This class selects the individuals to survive for the next generation

    eliminate_duplicates : bool
        If this flag is set no duplicates are allowed in the population (mostly likely only used for binary or discrete)

    verbose : int
        If larger than zero output is provided. (verbose=1 means some output, verbose=2 details for debugging)

    callback : func
        A callback function can be passed that is executed every generation. The parameters for the function
        are the algorithm itself, the number of evaluations so far and the current population.

            def callback(algorithm, n_evals, pop):
                print()

    """

    def __init__(self,
                 pop_size,
                 sampling,
                 selection,
                 crossover,
                 mutation,
                 survival,
                 n_offsprings=None,
                 eliminate_duplicates=False,
                 **kwargs
                 ):

        super().__init__(**kwargs)

        self.pop_size = pop_size
        self.sampling = sampling
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation
        self.survival = survival
        self.eliminate_duplicates = eliminate_duplicates
        self.n_offsprings = n_offsprings

        # default set the number of offsprings to the population size
        if self.n_offsprings is None:
            self.n_offsprings = pop_size

    def _solve(self, problem, evaluator):

        # create the population according to the factoring strategy
        pop = Population()
        if isinstance(self.sampling, np.ndarray):
            pop.X = self.sampling
        else:
            pop.X = self.sampling.sample(problem, self.pop_size, self)
        pop.F, pop.G = evaluator.eval(problem, pop.X)
        pop = self.survival.do(pop, self.pop_size, self)

        # setup initial generation
        n_gen = 0

        # while there are functions evaluations left
        while evaluator.has_next():

            # increase the generation and do printing and callback
            self._do_each_generation(n_gen, evaluator, pop)
            n_gen += 1

            # initialize selection and offspring methods
            off = Population()
            off.X = np.full((self.n_offsprings, problem.n_var), np.inf)
            self.selection.set_population(pop, self)

            n_off = 0
            n_parents = self.crossover.n_parents

            while n_off < self.n_offsprings:
                parents = self.selection.next(n_parents)
                X = self.crossover.do(problem, pop.X[parents, :])

                # if more offsprings than necessary - truncate them
                if X.shape[0] > self.n_offsprings - n_off:
                    X = X[:self.n_offsprings - n_off, :]

                off.X[n_off:n_off+X.shape[0], :] = X
                n_off = n_off + X.shape[0]

            off.X = self.mutation.do(problem, off.X)
            off.F, off.G = evaluator.eval(problem, off.X)

            # merge the population
            pop.merge(off)

            # eliminate all duplicates in the population
            if self.eliminate_duplicates:
                # pop.filter(unique_rows(pop.F))
                pop.filter(unique_rows(pop.X))

            # truncate the population
            pop = self.survival.do(pop, self.pop_size, self)

        self._do_each_generation(n_gen, evaluator, pop)
        return pop.X, pop.F, pop.G

    def _do_each_generation(self, n_gen, evaluator, pop):
        if self.verbose > 0:
            print('gen = %d' % (n_gen + 1))
        if self.verbose > 1:
            pass
        if self.callback is not None:
            self.callback(self, evaluator.counter, pop)

        if self.history is not None:
            self.history.append(
                {'n_gen': n_gen,
                 'n_evals': evaluator.counter,
                 'X': np.copy(pop.X),
                 'F': np.copy(pop.F)
                 })





from abc import abstractmethod

import numpy
import pymoo
from pymoo.model.evaluator import Evaluator
from pymoo.rand import random
from pymoo.util.non_dominated_rank import NonDominatedRank
from pymop.problem import Problem


class Algorithm:
    """

    This class represents the abstract class for any algorithm to be implemented. Most importantly it
    provides the solve method that is used to optimize a given problem.

    """

    def __init__(self,
                 verbose=False,
                 callback=None
                 ):
        self.verbose = verbose
        self.callback = callback


    def solve(self,
              problem,
              evaluator,
              seed=1,
              return_only_feasible=True,
              return_only_non_dominated=True,
              history=None,
              ):
        """

        Solve a given problem by a given evaluator. The evaluator determines the termination condition and
        can either have a maximum budget, hypervolume or whatever. The problem can be any problem the algorithm
        is able to solve.

        Parameters
        ----------

        problem: class
            Problem to be solved by the algorithm

        evaluator: class
            object that evaluates and saves the number of evaluations and determines the stopping condition

        seed: int
            Random seed for this run. Before the algorithm starts this seed is set.

        return_only_feasible:
            If true, only feasible solutions are returned.

        return_only_non_dominated
            If true, only the non dominated solutions are returned. Otherwise, it might be - dependend on the
            algorithm - the final population

        Returns
        -------
        X: matrix
            Design space

        F: matrix
            Objective space

        G: matrix
            Constraint space

        """

        # set the random seed for generator
        pymoo.rand.random.seed(seed)

        # just to be sure also for the others
        seed = pymoo.rand.random.randint(0, 100000)
        random.seed(seed)
        numpy.random.seed(seed)

        # add the history object
        self.history = history
        if self.history is True:
            self.history = []

        if not isinstance(evaluator, Evaluator):
            evaluator = Evaluator(evaluator)

        # run the initialization method first
        self._initialize(problem)

        # call the algorithm to solve the problem
        X, F, G = self._solve(problem, evaluator)

        if return_only_feasible:
            if G is not None and G.shape[0] == len(F) and G.shape[1] > 0:
                cv = Problem.calc_constraint_violation(G)
                X = X[cv <= 0, :]
                F = F[cv <= 0, :]
                if G is not None:
                    G = G[cv <= 0, :]

        if return_only_non_dominated:
            idx_non_dom = NonDominatedRank.calc_as_fronts(F, G)[0]
            X = X[idx_non_dom, :]
            F = F[idx_non_dom, :]
            if G is not None:
                G = G[idx_non_dom, :]

        return X, F, G

    @abstractmethod
    def _solve(self, problem, evaluator):
        pass

    def _initialize(self, problem):
        pass



from pymoo.model.selection import Selection
from pymoo.rand import random


class RandomSelection(Selection):
    def __init__(self):
        self.perm = None
        self.counter = None
        self.pop = None

    def set_population(self, pop, data):
        self.pop = pop

    def next(self, n_selected):
        if self.perm is None or self.counter + n_selected >= self.pop.size():
            self.perm = random.perm(self.pop.size())
            self.counter = 0

        selected = self.perm[self.counter:self.counter+n_selected]
        self.counter = self.counter + n_selected

        return selected



import numpy as np

from pymoo.model.crossover import Crossover
from pymoo.rand import random


class DifferentialEvolutionCrossover(Crossover):

    def __init__(self, variant="DE/rand/1", prob=0.5, weight=0.8, method="binomial", bounce_back_in_bounds=True):
        super().__init__(3, 1)
        self.prob = prob
        self.weight = weight
        self.type = method
        self.variant = variant
        self.bounce_back_in_bounds = bounce_back_in_bounds

    def _do(self, p, parents, children, X=None, best_i=None):

        if X is None:
            raise Exception("Please provide the current population design space to do the crossover.")

        n_var = X.shape[1]

        # do the crossover
        if self.type == "binomial":

            # uniformly for each individual and each entry
            r = random.random(size=X.shape) < self.prob

        elif self.type == "exponential":

            r = np.full(X.shape, False)

            # start point of crossover
            n = random.randint(0, n_var, size=X.shape[0])
            # length of chromosome to do the crossover
            L = np.argmax((random.random(X.shape) > self.prob), axis=1)

            # create for each individual the crossover range
            for i in range(X.shape[0]):
                for l in range(L[i] + 1):
                    r[i, (n[i] + l) % n_var] = True

        else:
            raise Exception("Unknown crossover type. Either binomial or exponential.")

        # the so called donor vector
        children = np.copy(X)

        if self.variant == "DE/rand/1":
            trial = parents[:, 2] + self.weight * (parents[:, 0] - parents[:, 1])
        elif self.variant == "DE/target-to-best/1":
            trial = X + self.weight * (X[best_i, :] - X) + self.weight * (parents[:, 0] - parents[:, 1])
        elif self.variant == "DE/best/1":
            trial = X[best_i, :] + (parents[:, 0] - parents[:, 1]) * self.weight
        else:
            raise Exception("DE variant %s not known." % self.variant)

        # set only if larger than F_CR
        children[r] = trial[r]

        # bounce back into bounds
        if self.bounce_back_in_bounds:
            smaller_than_min, larger_than_max = p.xl > children, p.xu < children
            children[smaller_than_min] = (p.xl + (p.xl - children))[smaller_than_min]
            children[larger_than_max] = (p.xu - (children - p.xu))[larger_than_max]

        return children



import numpy as np

from pymoo.model.algorithm import Algorithm
from pymoo.operators.crossover.real_differental_evolution_crossover import DifferentialEvolutionCrossover
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.rand import random
from pymoo.util.misc import parameter_less_constraints


class DifferentialEvolution(Algorithm):
    def __init__(self,
                 pop_size=100,
                 sampling=RealRandomSampling(),
                 crossover=DifferentialEvolutionCrossover(prob=0.5, weight=0.75, variant="DE/best/1",
                                                          method="binomial"),
                 **kwargs):

        super().__init__(**kwargs)

        self.pop_size = pop_size
        self.sampling = sampling
        self.crossover = crossover
        self.best_i = None

    def _solve(self, problem, evaluator):

        # create the population according to the factoring strategy
        if isinstance(self.sampling, np.ndarray):
            X = self.sampling
        else:
            X = self.sampling.sample(problem, self.pop_size, self)

        # evaluate and find ideal point
        F, CV = evaluator.eval(problem, X)
        F = parameter_less_constraints(F, CV)
        self.best_i = np.argmin(F)

        while evaluator.has_next():

            P = np.reshape(np.concatenate([random.perm(self.pop_size) for _ in range(self.crossover.n_parents)]),
                           (-1, self.crossover.n_parents))
            off_X = self.crossover.do(problem, X[P, :], X=X, best_i=self.best_i)

            # evaluate the offsprings
            off_F, off_CV = evaluator.eval(problem, off_X, return_constraints=2)
            off_F = parameter_less_constraints(off_F, off_CV)

            # replace if better
            off_is_better = np.where(off_F < F)[0]
            F[off_is_better, :] = off_F[off_is_better, :]
            X[off_is_better, :] = off_X[off_is_better, :]

            # find the index of the best individual
            self.best_i = np.argmin(F)

            if self.verbose:
                print(evaluator.counter, F[self.best_i, :])

        return X, F, np.zeros(self.pop_size)[:, None]



from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.default_operators import set_default_if_none, set_if_none
from pymoo.operators.survival.reference_line_survival import ReferenceLineSurvival
from pymoo.rand import random
from pymop.util import get_weights



class NSGAIII(GeneticAlgorithm):
    def __init__(self, var_type, ref_dirs=None, **kwargs):
        self.ref_dirs = ref_dirs
        set_default_if_none(var_type, kwargs)
        set_if_none(kwargs, 'survival', None)
        super().__init__(**kwargs)

    def _initialize(self, problem):
        super()._initialize(problem)

        # if survival not define differently
        if self.survival is None:

            # if ref dirs are not initialized do it based on the population size
            if self.ref_dirs is None:
                self.ref_dirs = get_weights(self.pop_size, problem.n_obj, func_random=random, method="uniform")

            # set the survival method itself
            self.survival = ReferenceLineSurvival(self.ref_dirs)




import random
import numpy as np

class SecureRandomGenerator:

    def random(self, size=None):
        if size is None:
            return random.SystemRandom().random()
        else:
            n = size[0]
            m = size[1]
            val = np.zeros((n, m))
            for i in range(n):
                for j in range(m):
                    val[i, j] = random.SystemRandom().random()

            return val


    def permutation(self, n):
        perm = list(range(n))
        random.SystemRandom().shuffle(perm)
        return perm

    def seed(self, x):
        pass



import numpy as np

from pymoo.util.dominator import Dominator


class NonDominatedRank:
    def __init__(self):
        pass

    @staticmethod
    def is_dominating(f, F):
        smaller = np.any(f < F, axis=1)
        larger = np.any(f > F, axis=1)
        dom = np.logical_and(smaller, np.logical_not(larger)) * 1 \
              + np.logical_and(larger, np.logical_not(smaller)) * -1
        return dom

    @staticmethod
    def get_non_dominated(F, F_, return_index=False):
        F_all = np.concatenate([F, F_], axis=0)
        front = NonDominatedRank().get_front(F_all)
        front = [i-F.shape[0] for i in front if i >= F.shape[0]]
        if return_index:
            return front
        return F_[front, :]

    @staticmethod
    def calc(F):
        fronts = NonDominatedRank.calc_as_fronts(F, None)
        return NonDominatedRank.calc_from_fronts(fronts)

    @staticmethod
    def calc_from_fronts(fronts):
        n = sum([len(f) for f in fronts])
        rank = np.zeros(n, dtype=np.int)
        for i in range(len(fronts)):
            for idx in fronts[i]:
                rank[idx] = i
        return rank

    @staticmethod
    def calc_as_fronts_naive(F, G):

        M = Dominator.calc_domination_matrix(F, G)
        fronts = []

        remaining = set(range(F.shape[0]))

        while len(remaining) > 0:

            front = []

            for i in remaining:

                is_dominated = False
                dominating = set()

                for j in front:
                    rel = M[i, j]
                    if rel == 1:
                        dominating.add(j)
                    elif rel == -1:
                        is_dominated = True
                        break

                if is_dominated:
                    continue
                else:
                    front = [x for x in front if x not in dominating]
                    front.append(i)

            [remaining.remove(e) for e in front]
            fronts.append(front)

        return fronts

    @staticmethod
    def get_front(F, G=None):
        return NonDominatedRank.calc_as_fronts(F, G, only_pareto_front=True)

    @staticmethod
    def calc_as_fronts(F, G=None, only_pareto_front=False):

        # calculate the dominance matrix
        n = F.shape[0]
        M = Dominator.calc_domination_matrix(F, G)

        fronts = []

        # final rank that will be returned
        n_ranked = 0
        ranked = np.zeros(n, dtype=np.int)

        # for each individual a list of all individuals that are dominated by this one
        is_dominating = [[] for _ in range(n)]

        # storage for the number of solutions dominated this one
        n_dominated = np.zeros(n)

        current_front = []

        for i in range(n):

            for j in range(i + 1, n):
                rel = M[i, j]
                if rel == 1:
                    is_dominating[i].append(j)
                    n_dominated[j] += 1
                elif rel == -1:
                    is_dominating[j].append(i)
                    n_dominated[i] += 1

            if n_dominated[i] == 0:
                current_front.append(i)
                ranked[i] = 1.0
                n_ranked += 1

        if only_pareto_front:
            return current_front

        # append the first front to the current front
        fronts.append(current_front)

        # while not all solutions are assigned to a pareto front
        while n_ranked < n:

            next_front = []

            # for each individual in the current front
            for i in current_front:

                # all solutions that are dominated by this individuals
                for j in is_dominating[i]:
                    n_dominated[j] -= 1
                    if n_dominated[j] == 0:
                        next_front.append(j)
                        ranked[j] = 1.0
                        n_ranked += 1

            fronts.append(next_front)
            current_front = next_front

        return fronts



import numpy as np

from pymoo.model.survival import Survival
from pymop.problem import Problem


class FitnessSurvival(Survival):
    """

    This survival method is just for single-objective algorithm.
    Simply sort by first constraint violation and then fitness value and truncate the worst individuals.


    """
    def _do(self, pop, size, data):

        if pop.F.shape[1] != 1:
            raise ValueError("FitnessSurvival can only used for single objective problems!")

        if pop.G is None or len(pop.G) == 0:
            CV = np.zeros(pop.F.shape[0])
        else:
            CV = Problem.calc_constraint_violation(pop.G)
            CV[CV < 0] = 0.0

        # sort by cv and fitness
        sorted_idx = sorted(range(pop.size()), key=lambda x: (CV[x], pop.F[x]))

        # now truncate the population
        sorted_idx = sorted_idx[:size]
        pop.filter(sorted_idx)

        return pop




from pymoo.algorithms.genetic_algorithm import GeneticAlgorithm
from pymoo.operators.default_operators import set_default_if_none, set_if_none
from pymoo.operators.survival.rnsga_reference_line_survival import ReferenceLineSurvival
import numpy as np

class ERNSGAII(GeneticAlgorithm):
    """
    Class Extended RNSGA-II
    """
    def __init__(self, var_type, ref_points=None, epsilon=0.001, weights=None, **kwargs):
        self.ref_dirs = np.array(ref_points)
        self.epsilon = epsilon
        self.weights = weights
        set_default_if_none(var_type, kwargs)
        set_if_none(kwargs, 'survival', None)
        super().__init__(**kwargs)

    def _initialize(self, problem):
        super()._initialize(problem)

        # if survival not define differently
        if self.survival is None:

            # set the survival method itself
            self.survival = ReferenceLineSurvival(self.ref_dirs, self.epsilon, self.weights)



import numpy as np

from pymoo.rand.random_generator import RandomGenerator


class MyRandomGenerator(RandomGenerator):

    def __init__(self):
        self._seed = -1
        self.oldrand = np.zeros(55)
        self.jrand = 0

    def seed(self, s):
        import random
        random.seed(s)
        self._seed = random.random()

        self.__warmup_random()

    def rand(self, size=None):

        if isinstance(size, tuple) and len(size) == 1:
            size = size[0]

        if size is None:
            return self.__randomperc()
        elif isinstance(size, int):
            val = np.zeros(size)
            for j in range(size):
                val[j] = self.__randomperc()
            return val
        elif isinstance(size, tuple) and len(size) == 2:

            n = size[0]
            m = size[1]
            val = np.zeros((n, m))
            for i in range(n):
                for j in range(m):
                    val[i, j] = self.__randomperc()

            return val
        else:
            raise Exception("Random Generator size % is not provided." % size)

    def perm(self, n):
        a = np.array(range(n))
        for i in range(n):
            rand = self.randint(i, n - 1)
            temp = a[rand]
            a[rand] = a[i]
            a[i] = temp
        return a

    def randint(self, low, high, size=None):
        val = self.rand(size=size)
        return (low + (val * (high - low))).astype(np.int)

    def __warmup_random(self):
        self.oldrand[54] = self._seed
        new_random = 0.000000001
        prev_random = self._seed
        for j1 in range(1, 55):
            ii = (21 * j1) % 54
            self.oldrand[ii] = new_random
            new_random = prev_random - new_random
            if new_random < 0.0:
                new_random += 1.0
            prev_random = self.oldrand[ii]

        self.__advance_random()
        self.__advance_random()
        self.__advance_random()
        self.jrand = 0

    def __randomperc(self):
        self.jrand += 1
        if self.jrand >= 55:
            self.jrand = 1
            self.__advance_random()
        return self.oldrand[self.jrand]

    def __advance_random(self):
        for j1 in range(24):
            new_random = self.oldrand[j1] - self.oldrand[j1 + 31]
            if new_random < 0.0:
                new_random = new_random + 1.0
            self.oldrand[j1] = new_random

        for j1 in range(24, 55):
            new_random = self.oldrand[j1] - self.oldrand[j1 - 24]
            if new_random < 0.0:
                new_random = new_random + 1.0
            self.oldrand[j1] = new_random




import random
import numpy as np


class DefaultRandomGenerator:

    def rand(self, size=None):
        if size is None:
            return random.random()
        else:
            n = size[0]
            m = size[1]
            val = np.zeros((n, m))
            for i in range(n):
                for j in range(m):
                    val[i, j] = random.random()

            return val


    def perm(self, length, n=1):
        perms = []
        for _ in range(n):
            perm = list(range(length))
            random.shuffle(perm)
            perms.append(perm)
        return perms

    def seed(self, x):
        random.seed(x)

    def randint(self, low, high, size=None):
        if low >= high:
            return low
        else:
            res = low + (self.rand(size) * (high - low))
            if res > high:
                res = high
        return int(res)




import numpy as np

from pymoo.rand.random_generator import RandomGenerator


class NumpyRandomGenerator(RandomGenerator):

    def seed(self, x):
        np.random.seed(x)

    def rand(self, size=None):
        if size is None:
            return np.random.random()
        elif isinstance(size, int):
            return np.random.random(size)
        else:
            return np.random.random((size[0], size[1]))

    def randint(self, low, high, size=None):
        return np.random.randint(low, high, size)

    def perm(self, n):
        return np.random.permutation(n)




from pymoo.model.crossover import Crossover
from pymoo.rand import random


class BinaryUniformCrossover(Crossover):
    def __init__(self):
        super().__init__(2, 2)

    def _do(self, p, parents, children, data=None):

        for i in range(p.n_var):

            if random.random() < 0.5:
                children[0,i] = parents[0,i]
                children[1, i] = parents[1, i]
            else:
                children[0, i] = parents[1, i]
                children[1, i] = parents[0, i]

        return children



import numpy as np


class Population:
    def __init__(self, X=None, F=None, G=None):
        self.X = X  # design variables
        self.F = F  # objective values
        self.G = G  # constraint violations as vectors

    def merge(self, other):
        if self.X is not None:
            self.X = np.concatenate([self.X, other.X])
        else:
            self.X = other.X

        if self.F is not None:
            self.F = np.concatenate([self.F, other.F])
        else:
            self.F = other.F

        if self.G is not None and other.G is not None:
            self.G = np.concatenate([self.G, other.G])
        else:
            self.G = other.G

        return self

    def size(self):
        if self.F is None:
            return 0
        else:
            return len(self.F)

    def filter(self, v):
        if self.X is None:
            return 0
        else:
            self.X = self.X[v]
            self.F = self.F[v]
            if self.G is not None:
                self.G = self.G[v]
        return self



import numpy as np
from pymoo.model.indicator import Indicator


def hypervolume(F):
    pass

class Hypervolume(Indicator):
    def __init__(self, reference_point):
        Indicator.__init__(self)
        self.reference_point = reference_point

    def _calc(self, F):

        if F.shape[1] == 0:
            return 0.0

        F = np.array([e for e in F if np.all(e < self.reference_point)])
        if len(F) == 0:
            return 0.0

        hv = hypervolume(F)
        return hv.compute(self.reference_point)



import copy
import numpy as np
from scipy import special


def get_ref_dirs_from_section(n_obj, n_sections):

    if n_obj == 1:
        return np.array([1.0])

    # all possible values for the vector
    sections = np.linspace(0, 1, num=n_sections + 1)[::-1]

    ref_dirs = []
    ref_recursive([], sections, 0, n_obj, ref_dirs)
    return np.array(ref_dirs)


# returns the closest possible number of references lines to given one
def get_ref_dirs_from_n(n_obj, n_refs, max_sections=100):
    n_sections = np.array([get_number_of_reference_directions(n_obj, i) for i in range(max_sections)])
    idx = np.argmin((n_sections < n_refs).astype(np.int))
    M = get_ref_dirs_from_section(n_obj, idx-1)
    M[M==0] = 0.000001
    return M


def ref_recursive(v, sections, level, max_level, result):
    v_sum = np.sum(np.array(v))

    # sum slightly above or below because of numerical issues
    if v_sum > 1.0001:
        return
    elif level == max_level:
        if 1.0 - v_sum < 0.0001:
            result.append(v)
    else:
        for e in sections:
            next = list(v)
            next.append(e)
            ref_recursive(next, sections, level + 1, max_level, result)


def get_number_of_reference_directions(n_obj, n_sections):
    return int(special.binom(n_obj + n_sections - 1, n_sections))


def get_ref_dirs_from_points(points, n_obj):

    extreme_reference_points = np.eye(n_obj)

    ref_points = np.append(points, extreme_reference_points, axis=0)
    return ref_points

def get_ref_dirs_from_points(ref_point, n_obj, alpha=0.1):
    """
    This function takes user specified reference points, and creates smaller sets of equidistant
    Das-Dennis points around the projection of user points on the Das-Dennis hyperplane
    :param ref_point: List of user specified reference points
    :param n_obj: Number of objectives to consider
    :param alpha: Shrinkage factor (0-1), Smaller = tigher convergence, Larger= larger convergence
    :return: Set of reference points
    """
    ref_dirs = []
    n_vector = np.ones(n_obj) / np.linalg.norm(np.ones(n_obj))  # Normal vector of Das Dennis plane
    point_on_plane = np.eye(n_obj)[0]  # Point on Das-Dennis
    reference_directions = get_ref_dirs_from_n(n_obj, 21)  # Das-Dennis points

    for point in ref_point:
        # ref_proj = point - np.dot(point - point_on_plane, n_vector) * n_vector
        # TODO: Compute which is faster, a copy.deepcopy, or recomputing all the points from get_ref_dirs_from_n
        ref_dir = copy.deepcopy(reference_directions)  # Copy of computed reference directions
        for i in range(n_obj):  # Shrink Das-Dennis points by a factor of alpha
            ref_dir[:, i] = point[i] + alpha * (ref_dir[:, i] - point[i])
        for d in ref_dir:  # Project shrunked Das-Dennis points back onto original Das-Dennis hyperplane
            ref_dirs.append(d - np.dot(d - point_on_plane, n_vector) * n_vector)
    # TODO: Extreme points are only extreme of the scale is normalized between 0-1, how to make them truly extreme?
    ref_dirs.extend(np.eye(n_obj))  # Add extreme points
    return np.array(ref_dirs)


if __name__ == '__main__':

    test = get_ref_dirs_from_n(2, 100)

    # for i in [3]:
    #     for j in range(20):
    #         test = get_ref_dirs_from_section(i, j)
    #         print(j, len(test), get_number_of_reference_directions(i, j))
    # print()
    import pylab as pl
    fig = pl.subplot()
    pl.scatter(test[:, 0], test[:, 1])



from abc import abstractmethod

import numpy as np


class Crossover:
    """

    This class is the base class for a crossover. In general any crossover can be used.
    It is up to the algorithm composer to take care of not using a crossover for real values in a binary problem.
    Each crossover needs to define the number of parents that are needed and also the number of children
    that are composed.

    """

    def __init__(self, n_parents, n_children):
        self.n_parents = n_parents
        self.n_children = n_children

    def do(self, problem, parents, **kwargs):
        """

        This method executes the crossover on the parents. This class wraps the implementation of the class
        that implements the crossover.


        Parameters
        ----------
        problem: class
            The problem to be solved. Provides information such as lower and upper bounds or feasibility
            conditions for custom crossovers.

        parents: matrix
            The parents as a matrix. Each row is a parent and the columns are equal to the parameter length.

        data:
            Any additional data that might be necessary to perform the crossover. E.g. constants of an algorithm.

        Returns
        -------
        children: matrix
            The children as a matrix. n_children rows and the number of columns is equal to the variable
            length of the problem.

        """
        n_var = parents.shape[1]

        #if n_var != problem.n_var:
        #    raise ValueError('Exception during crossover: Variable length is not equal to the defined one in problem.')

        #if self.n_parents != parents.shape[0]:
        #    raise ValueError('Exception during crossover: Number of parents differs from defined at crossover.')

        children = np.full((self.n_children, n_var), np.inf)
        return self._do(problem, parents, children,  **kwargs)


    @abstractmethod
    def _do(self, problem, parents, children,  **kwargs):
        pass



from abc import abstractmethod


class RandomGenerator:
    @abstractmethod
    def seed(self, s):
        pass

    @abstractmethod
    def perm(self, size):
        pass

    @abstractmethod
    def rand(self, size=None):
        pass

    @abstractmethod
    def randint(self, low, high, size=None):
        pass


from pymoo.configuration import Configuration


class Singleton:
    __instance = None

    @staticmethod
    def get_instance():
        if Singleton.__instance is None:
            Singleton.__instance = Configuration.rand
            Singleton.__instance.seed(1)
        return Singleton.__instance


def seed(s):
    return Singleton.get_instance().seed(s)


def perm(size):
    return Singleton.get_instance().perm(size)


def random(*params, size=None):
    if len(params) > 0:
        return Singleton.get_instance().rand(size=params)
    else:
        return Singleton.get_instance().rand(size=size)


def randint(low, high, size=None):
    return Singleton.get_instance().randint(low, high, size=size)



from abc import abstractmethod

import numpy as np


class Mutation:
    def do(self, problem, X):
        Y = np.full(X.shape, np.inf)
        return self._do(problem, X, Y)

    @abstractmethod
    def _do(self, problem, X):
        pass



import os
import sys
import time

import numpy as np
from pymop.zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6
from pymop.dtlz import DTLZ2

from pymoo.algorithms.ERNSGAII import RNSGAIII
from pymoo.model.evaluator import Evaluator
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover

from pymoo.indicators.rmetric import RMetric

n_runs = 30
output = os.path.join('.', 'benchmark/')
if not os.path.exists(output):
    os.makedirs(os.getcwd()+'/benchmark/')

def ZDT_Test():
    ref_dirs = {ZDT1: [[(0.2, 0.4), (0.8, 0.4)],
                         [(0.2, 0.6), (0.4, 0.6), (0.5, 0.2), (0.7, 0.2), (0.9, 0)]],
                ZDT2: [[(0.2, 0.8), (0.7, 1), (0.8, 0.2)]],
                ZDT3: [[(0.1, 0.6), (0.3, 0.2), (0.7, -0.25)]]}
    crossover = SimulatedBinaryCrossover(10)

    p = []
    name_list = ["ZDT1_1", "ZDT1_2", "ZDT2_1", "ZDT3_1"]
    for problem, ref_points in ref_dirs.items():
        for i, points in enumerate(ref_points):

            sublist = []
            for algorithm in [RNSGAIII("real", pop_size=100, ep=0.001, crossover=crossover, ref_dirs=points, verbose=0)]:
                for run in range(1, n_runs + 1):
                    name = problem.__class__.__name__ + '_' + str(i) + '_' + str(run)
                    sublist.append((algorithm, problem, run, points, name))
                    # yield (algorithm, problem, run)
            p.append(sublist)
    return p, name_list

def DTLZ_test():
    ref_dirs = {DTLZ2: [{"n_var": 11, "n_obj": 3, "ref_points": [[[0.2, 0.2, 0.6], [0.8, 0.8, 0.8]]]},
                          {"n_var": 14, "n_obj": 5, "ref_points": [[[0.5, 0.5, 0.5, 0.5, 0.5], [0.2, 0.2, 0.2, 0.2, 0.8]]]},
                          {"n_var": 19, "n_obj": 10, "ref_points": [[[0.25 for i in range(10)]]]}
                          ]
                }
    crossover = SimulatedBinaryCrossover(10)
    p = []
    name_list = ["DTLZ2_1", "DTLZ2_2", "DTLZ2_3"]
    for problem, setup in ref_dirs.items():
        for params in setup:
            parameters = {"n_var": params["n_var"], "n_obj": params["n_obj"]}
            prob = problem(**parameters)
            for i, points in enumerate(params["ref_points"]):
                sublist = []
                for algorithm in [RNSGAIII("real", pop_size=100, ep=0.01, crossover=crossover, ref_dirs=points, verbose=0)]:
                    for run in range(1, n_runs + 1):
                        name = problem.__class__.__name__ + '_' + str(i) + '_' + str(run)
                        sublist.append((algorithm, prob, run, points, name))
                        # yield (algorithm, prob, run)
            p.append(sublist)
    return p, name_list

if __name__ == '__main__':
    param1, name1 = ZDT_Test()
    param2, name2 = DTLZ_test()

    params = param1 + param2
    names = name1 + name2
    n_gen = 500

    results = []

    id = int(sys.argv[1])
    print("Run for parameter settings: %s" % (id))

    run = params[id]
    name = names[id]
    output = output + name
    if not os.path.exists(output):
        os.makedirs(output)

        print(output)

    i = 0
    for algorithm, problem, run, points, n in run:
        try:
            problem = problem()
        except TypeError:
            pass
        points = np.array(points)
        eval = Evaluator(100 * n_gen)
        start_time = time.time()
        X, F, G = algorithm.solve(problem, evaluator=eval, seed=run)
        print("--- %s seconds ---" % (time.time() - start_time))
        fname = os.path.join(output, problem.name()+"_"+str(i))
        print(fname)
        np.savetxt(fname + ".out", F)
        if id <= 5:
            metric_time = time.time()
            igd, hv = RMetric(curr_pop=F, whole_pop=F, ref_points=points, problem=problem).calc()
            results.append(np.array([run, igd, hv]))
            print(igd, " - ", hv)
            print("--- %s seconds ---" % (time.time() - metric_time))
        break
        i += 1

    results = np.array(results)
    if id <=5:
        np.savetxt(os.path.join(output, "results.out"), results, fmt=("%-3i","%-10.7f","%-10.7f"))
    np.savetxt(os.path.join(output, "points.out"), points, fmt="%-5.2f")


import numpy as np
from scipy.spatial.distance import cdist

from pymoo.model.survival import Survival
from pymoo.rand import random
from pymoo.util.misc import normalize_by_asf_interceptions
from pymoo.util.non_dominated_rank import NonDominatedRank


class ReferenceLineSurvival(Survival):
    def __init__(self, ref_dirs):
        super().__init__()
        self.ref_dirs = ref_dirs

    def _do(self, pop, n_survive, data, return_only_index=False):

        fronts = NonDominatedRank.calc_as_fronts(pop.F, pop.G)

        # all indices to survive
        survival = []

        for front in fronts:
            if len(survival) + len(front) > n_survive:
                break
            survival.extend(front)

        # filter the front to only relevant entries
        pop.filter(survival + front)
        survival = list(range(0, len(survival)))
        last_front = np.arange(len(survival), pop.size())

        N = normalize_by_asf_interceptions(pop.F, return_bounds=False)
        # N = normalize(pop.F, np.zeros(pop.F.shape[1]), np.ones(pop.F.shape[1]))

        # if the last front needs to be splitted
        n_remaining = n_survive - len(survival)
        if n_remaining > 0:

            dist_matrix = calc_perpendicular_dist_matrix(N, self.ref_dirs)
            niche_of_individuals = np.argmin(dist_matrix, axis=1)
            dist_to_niche = dist_matrix[np.arange(len(dist_matrix)), niche_of_individuals]

            # for each reference direction the niche count
            niche_count = np.zeros(len(self.ref_dirs))
            for i in niche_of_individuals[survival]:
                niche_count[i] += 1

            # relative index to dist and the niches just of the last front
            dist_to_niche = dist_to_niche[last_front]
            niche_of_individuals = niche_of_individuals[last_front]

            # boolean array of elements that are considered for each iteration
            remaining_last_front = np.full(len(last_front), True)

            while n_remaining > 0:

                # all niches where new individuals can be assigned to
                next_niches_list = np.unique(niche_of_individuals[remaining_last_front])

                # pick a niche with minimum assigned individuals - break tie if necessary
                next_niche_count = niche_count[next_niches_list]
                next_niche = np.where(next_niche_count == next_niche_count.min())[0]
                next_niche = next_niche[random.randint(0, len(next_niche))]
                next_niche = next_niches_list[next_niche]

                # indices of individuals that are considered and assign to next_niche
                next_ind = np.where(np.logical_and(niche_of_individuals == next_niche, remaining_last_front))[0]

                if len(next_ind) == 1:
                    next_ind = next_ind[0]
                elif niche_count[next_niche] == 0:
                    next_ind = next_ind[np.argmin(dist_to_niche[next_ind])]
                else:
                    next_ind = next_ind[random.randint(0, len(next_ind))]

                remaining_last_front[next_ind] = False
                survival.append(last_front[next_ind])

                niche_count[next_niche] += 1
                n_remaining -= 1

        if return_only_index:
            return survival

        # now truncate the population
        pop.filter(survival)

        return pop


def calc_perpendicular_dist(v, u):
    norm_u = np.linalg.norm(u)
    scalar_proj = np.dot(v, u) / norm_u
    proj = scalar_proj * u / norm_u
    return np.linalg.norm(proj - v)


def calc_perpendicular_dist_matrix_slow(N, ref_dirs):
    return cdist(N, ref_dirs, metric=calc_perpendicular_dist)


def calc_perpendicular_dist_matrix(N, ref_dirs):
    u = np.tile(ref_dirs, (len(N), 1))
    v = np.repeat(N, len(ref_dirs), axis=0)

    norm_u = np.linalg.norm(u, axis=1)

    scalar_proj = np.sum(v * u, axis=1) / norm_u
    proj = scalar_proj[:, None] * u / norm_u[:, None]
    val = np.linalg.norm(proj - v, axis=1)
    matrix = np.reshape(val, (len(N), len(ref_dirs)))

    return matrix



from abc import abstractmethod


class Survival:
    """

    The class represents the survival selection during the evolution. Only the fittest can survive.

    """

    def do(self, pop, n_survive, data):
        """

        The whole population is provided and the number of individuals to survive. If the number of survivers
        is smaller than the populations a survival selection is done. Otherwise, the elements might only
        be sorted by a specific criteria. This depends on the survival implementation.

        Parameters
        ----------
        pop: class
            The population.
        n_survive: int
            number of individuals that should survive.
        data: class
            Any additional data that might be needed to choose what individuals survive.

        Returns
        -------
        pop: class
            The population that has survived.

        """
        return self._do(pop, n_survive, data)

    @abstractmethod
    def _do(self, pop, n_survive, data):
        pass



import os
import sys
import time

import numpy as np
from pymop.zdt import ZDT1, ZDT2, ZDT3, ZDT4, ZDT6
from pymop.dtlz import DTLZ2

from pymoo.algorithms.RNSGAIII import RNSGAIII
from pymoo.model.evaluator import Evaluator
from pymoo.operators.crossover.real_simulated_binary_crossover import SimulatedBinaryCrossover

n_runs = 30
output = os.path.join('..', 'benchmark', 'RNSGAII_Benchmark')
if not os.path.exists(output):
    output = os.path.join('.')


def ZDT_Test():
    ref_dirs = {ZDT1(): [[(0.2, 0.4), (0.8, 0.4)],
                         [(0.2, 0.6), (0.4, 0.6), (0.5, 0.2), (0.7, 0.2), (0.9, 0)]],
                ZDT2(): [[(0.2, 0.8), (0.7, 1), (0.8, 0.2)]],
                ZDT3(): [[(0.1, 0.6), (0.3, 0.2), (0.7, -0.25)]]}
    crossover = SimulatedBinaryCrossover(10)

    for problem, ref_points in ref_dirs.items():
        for points in ref_points:
            for algorithm in [RNSGAIII("real", pop_size=100, ep=0.001, crossover=crossover, ref_dirs=points, verbose=0)]:
                for run in range(1, n_runs + 1):
                    yield (algorithm, problem, run)

def DTLZ_test():
    ref_dirs = {DTLZ2: [{"n_var": 11, "n_obj": 3, "ref_points": [[[0.2, 0.2, 0.6], [0.8, 0.8, 0.8]]]},
                          {"n_var": 14, "n_obj": 5, "ref_points": [[[0.5, 0.5, 0.5, 0.5, 0.5], [0.2, 0.2, 0.2, 0.2, 0.8]]]},
                          {"n_var": 19, "n_obj": 10, "ref_points": [[[0.25 for i in range(10)]]]}
                          ]
                }
    crossover = SimulatedBinaryCrossover(10)

    for problem, setup in ref_dirs.items():
        for params in setup:
            parameters = {"n_var": params["n_var"], "n_obj": params["n_obj"]}
            print(problem)
            prob = problem(**parameters)
            for points in params["ref_points"]:
                for algorithm in [RNSGAIII("real", pop_size=100, ep=0.01, crossover=crossover, ref_dirs=points, verbose=0)]:
                    for run in range(1, n_runs + 1):
                        yield (algorithm, prob, run)


if __name__ == '__main__':

    params = list(ZDT_Test()) + list(DTLZ_test())
    total_time = time.time()
    for i in range(len(params)):

        algorithm, problem, run = params[i]
        n_gen = 500
        eval = Evaluator(100*n_gen)
        start_time = time.time()
        X, F, G = algorithm.solve(problem, evaluator=eval, seed=run)
        print("--- %s seconds ---" % (time.time() - start_time))

        # save the result as a test
        fname = os.path.join(output, 'pynsga3_' + problem.__class__.__name__ + '_%s' % run)
        np.savetxt(fname + ".out", F)
        print(fname + ".out")
        # save_hist(fname + ".hist", eval.data)
    print(f"Benchmark took {time.time()-total_time} seconds")


import setuptools
from setuptools import setup

__author__ = "Julian Blank"
__version__ = '0.1.2'

try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = ''



setup(
    name="pymoo",
    version=__version__,
    author=__author__,
    author_email="blankjul@egr.msu.edu",
    description="Multi-Objective Optimization Algorithms",
    long_description=long_description,
    url="https://github.com/julesy89/pymoo",
    license='MIT',
    keywords="optimization",
    packages=setuptools.find_packages(),
    install_requires=['pymop', 'numpy', 'scipy', 'matplotlib']
)



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.model.algorithm import Algorithm
from pymoo.operators.crossover.real_differental_evolution_crossover import DifferentialEvolutionCrossover
from pymoo.operators.mutation.real_polynomial_mutation import PolynomialMutation
from pymoo.operators.sampling.real_random_sampling import RealRandomSampling
from pymoo.rand import random
from pymop.util import get_weights


class MOEAD(Algorithm):
    def __init__(self,
                 pop_size=100,
                 n_neighbors=20,
                 sampling=RealRandomSampling(),
                 mutation=PolynomialMutation(),
                 crossover=DifferentialEvolutionCrossover(),
                 **kwargs):

        super().__init__(**kwargs)

        self.pop_size = pop_size
        self.sampling = sampling

        self.crossover = crossover
        self.mutation = mutation
        self.n_neighbors = n_neighbors

        # initialized when problem is known
        self.weights = None
        self.neighbours = None

    def _initialize(self, problem):
        self.weights = get_weights(self.pop_size, problem.n_obj, func_random=random.random, method="uniform")
        self.neighbours = np.argsort(cdist(self.weights, self.weights), axis=1)[:, :self.n_neighbors]

    def _solve(self, problem, evaluator):

        # create the population according to the factoring strategy
        if isinstance(self.sampling, np.ndarray):
            X = self.sampling
        else:
            X = self.sampling.sample(problem, self.pop_size, self)

        # evaluate and find ideal point
        F, _ = evaluator.eval(problem, X)
        ideal_point = np.min(F, axis=0)

        iteration = 1

        while evaluator.has_next():

            # iterate for each member of the population
            for i in range(self.pop_size):

                # select the parents from the neighbourhood
                parents = self.neighbours[i, random.perm(self.n_neighbors)[:self.crossover.n_parents]]

                # do recombination and create an offspring
                off_X = self.crossover.do(problem, X[None, parents,:], X=X[[i],:])

                # do the mutation
                off_X = self.mutation.do(problem, off_X)

                # evaluate the offspring
                off_F, _ = evaluator.eval(problem, off_X)

                # update the ideal point
                ideal_point = np.min(np.concatenate([ideal_point[None, :], off_F], axis=0), axis=0)

                # for each offspring that was created
                for k in range(self.crossover.n_children):

                    # the weights of each neighbor
                    weights_of_neighbors = self.weights[self.neighbours[i], :]

                    # calculate the decomposed values for each neighbour
                    FV = tchebi(F[self.neighbours[i]], weights_of_neighbors, ideal_point)
                    off_FV = tchebi(off_F[[k], :], weights_of_neighbors, ideal_point)

                    # get the absolute index in F where offspring is better than the current F (decomposed space)
                    off_is_better = self.neighbours[i][np.where(off_FV < FV)[0]]
                    F[off_is_better, :] = off_F[k, :]
                    X[off_is_better, :] = off_X[k, :]

            self._do_each_generation(iteration, evaluator, X, F)
            iteration += 1

        return X, F, np.zeros(self.pop_size)[:, None]

    def _do_each_generation(self, n_gen, evaluator, X, F):
        if self.verbose > 0:
            print('gen = %d' % (n_gen + 1))
        if self.verbose > 1:
            pass
        if self.callback is not None:
            pass
            #self.callback(self, evaluator.counter, pop)

        if self.history is not None:
            self.history.append(
                {'n_gen': n_gen,
                 'n_evals': evaluator.counter,
                 'X': np.copy(X),
                 'F': np.copy(F)
                 })


def tchebi(F, weights, ideal_point):
    v = np.abs((F - ideal_point) * weights)
    return np.max(v, axis=1)




import numpy as np
from pymoo.model.selection import Selection
from pymoo.rand import random


class TournamentSelection(Selection):
    """

      The Tournament selection is used to simulated a tournament between individuals. The pressure balances
      greedy the genetic algorithm will be.


    """

    def __init__(self, f_comp=None, pressure=2):
        """

        Parameters
        ----------
        f_comp: func
            The function to compare two individuals. It has the shape: comp(pop, indices, data) and returns the winner.
            If the function is None it is assumed the population is sorted by a criterium and only indices are compared.

        pressure: int
            The selection pressure to be applied. Default it is a binary tournament.
        """

        # selection pressure to be applied
        self.pressure = pressure

        self.f_comp = f_comp
        if self.f_comp is None:
            self.f_comp = self.select_by_min_index

        # attributes that will be set during the optimization
        self.pop = None
        self.perm = None
        self.counter = None
        self.data = None

    def set_population(self, pop, data):
        """

        Parameters
        ----------
        pop: class
            The population to be selected from.
        data: class
            Any additional data that might be needed for the selection.

        """
        self.pop = pop
        self.data = data
        self.perm = random.perm(self.pop.size())
        self.counter = 0

    def next(self, n_selected):
        """

        Parameters
        ----------
        n_selected: int
            Number of individuals to be selected.

        Returns
        -------
        v: vector
            Selected indices of individuals as a integer vector.

        """

        selected = np.zeros(n_selected, dtype=np.int)

        for i in range(n_selected):

            if self.counter + self.pressure >= self.pop.size():
                self.perm = random.perm(self.pop.size())
                self.counter = 0
            selected[i] = self.f_comp(self.pop, self.perm[self.counter:self.counter + self.pressure], self.data)
            self.counter = self.counter + self.pressure

        return selected

    # select simply by minimum index and assume sorting of the population according selection criteria
    def select_by_min_index(self, pop, indices, data):
        return np.min(indices)



import numpy as np

from pymoo.model.survival import Survival
from pymoo.util.misc import normalize
from pymoo.util.non_dominated_rank import NonDominatedRank


class RankAndCrowdingSurvival(Survival):
    def _do(self, pop, size, data, return_sorted_idx=False):

        fronts = NonDominatedRank.calc_as_fronts(pop.F, pop.G)
        rank = NonDominatedRank.calc_from_fronts(fronts)
        crowding = np.zeros(pop.F.shape[0])

        for front in fronts:
            cd_of_front = RankAndCrowdingSurvival.calc_crowding_distance(pop.F[front, :])
            crowding[front] = cd_of_front

        sorted_idx = sorted(range(pop.size()), key=lambda x: (rank[x], -crowding[x]))

        if return_sorted_idx:
            return sorted_idx

        # now truncate the population
        sorted_idx = sorted_idx[:size]
        pop.filter(sorted_idx)
        rank = rank[sorted_idx]
        crowding = crowding[sorted_idx]

        if data is not None:
            data.rank = rank
            data.crowding = crowding

        return pop

    @staticmethod
    def calc_crowding_distance(F, F_min=None, F_max=None):

        n = F.shape[0]
        m = F.shape[1]

        if n == 0:
            return []
        elif n == 1 or n == 2:
            return np.full(n, np.inf)
        else:

            cd = np.zeros(n)
            N = normalize(F, F_min, F_max)

            # for each objective
            for j in range(m):

                # sort by its objective
                sorted_idx = np.argsort(N[:, j])

                # set the corner points to infinity
                cd[sorted_idx[0]] = np.inf
                cd[sorted_idx[-1]] = np.inf

                # add up the crowding measure for all points in between
                for i in range(1, n - 1):
                    if np.isinf(cd[sorted_idx[i]]):
                        continue
                    else:
                        cd[sorted_idx[i]] += N[sorted_idx[i + 1], j] - N[sorted_idx[i - 1], j]

        return cd



import time

from matplotlib import animation
import matplotlib.pyplot as plt
import numpy as np

from pymoo.algorithms.so_DE import DifferentialEvolution
from pymop.rastrigin import Rastrigin

if __name__ == '__main__':

    # load the problem instance
    from pymop.zdt import ZDT1
    problem = ZDT1(n_var=30)
    problem = Rastrigin(n_var=30)

    # create the algorithm instance by specifying the intended parameters
    from pymoo.algorithms.NSGAII import NSGAII
    algorithm = NSGAII("real", pop_size=100, verbose=True)
    algorithm = DifferentialEvolution(pop_size=100, verbose=True)

    start_time = time.time()

    # save the history in an object to observe the convergence over generations
    history = []

    # number of generations to run it
    n_gen = 200

    # solve the problem and return the results
    X, F, G = algorithm.solve(problem,
                              evaluator=(100 * n_gen),
                              seed=2,
                              return_only_feasible=False,
                              return_only_non_dominated=False,
                              history=history)

    print("--- %s seconds ---" % (time.time() - start_time))

    scatter_plot = True
    save_animation = False

    # get the problem dimensionality
    is_2d = problem.n_obj == 2
    is_3d = problem.n_obj == 3

    if scatter_plot and is_2d:
        plt.scatter(F[:, 0], F[:, 1])
        plt.show()

    if scatter_plot and is_3d:
        fig = plt.figure()
        from mpl_toolkits.mplot3d import Axes3D
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(F[:, 0], F[:, 1], F[:, 2])
        plt.show()

    # create an animation to watch the convergence over time
    if is_2d and save_animation:

        fig = plt.figure()
        ax = plt.gca()

        _F = history[0]['F']
        pf = problem.pareto_front()
        plt.scatter(pf[:,0], pf[:,1], label='Pareto Front', s=60, facecolors='none', edgecolors='r')
        scat = plt.scatter(_F[:, 0], _F[:, 1])


        def update(frame_number):
            _F = history[frame_number]['F']
            scat.set_offsets(_F)

            # get the bounds for plotting and add padding
            min = np.min(_F, axis=0) - 0.1
            max = np.max(_F, axis=0) + 0.

            # set the scatter object with padding
            ax.set_xlim(min[0], max[0])
            ax.set_ylim(min[1], max[1])


        # create the animation
        ani = animation.FuncAnimation(fig, update, frames=range(n_gen))

        # write the file
        Writer = animation.writers['ffmpeg']
        writer = Writer(fps=6, bitrate=1800)
        ani.save('%s.mp4' % problem.name(), writer=writer)



from abc import abstractmethod


class Selection:
    """

    This class represents the selection of individuals for the mating.

    """

    @abstractmethod
    def set_population(self, pop, data):
        """

        Set the population to be selected from.

        Parameters
        ----------
        pop: class
            An object that stores the current population to be selected from.
        data: class
            Any other additional data that might be needed for the selection procedure.

        """
        pass

    @abstractmethod
    def next(self, n_select):
        """

        Choose from the population new individuals to be selected.

        Parameters
        ----------
        n_select: int
            How many individuals that should be selected.

        Returns
        -------

        select: vector
            A vector that contains the result indices of selected individuals.

        """

        pass



import numpy as np
from numpy.linalg import LinAlgError


def denormalize(x, x_min, x_max):
    return x * (x_max - x_min) + x_min

def normalize(x, x_min=None, x_max=None, return_bounds=False):
    if x_min is None:
        x_min = np.min(x, axis=0)
    if x_max is None:
        x_max = np.max(x, axis=0)

    denom = x_max - x_min
    denom += 1e-30

    res = (x - x_min) / denom
    if not return_bounds:
        return res
    else:
        return res, x_min, x_max


def normalize_by_asf_interceptions(x, return_bounds=False):
    # find the x_min point
    n_obj = x.shape[1]
    x_min = np.min(x, axis=0)
    # transform the objective that 0 means the best
    F = x - x_min

    # calculate the asf matrix
    asf = np.eye(n_obj)
    asf = asf + (asf == 0) * 1e-16

    # result matrix with the selected points
    S = np.zeros((n_obj, n_obj))

    # find for each direction the best
    for i in range(len(asf)):
        val = np.max(F / asf[i, :], axis=1) + 0.0001 * np.sum(F / asf[i, :])
        S[i, :] = F[np.argmin(val), :]

    try:
        b = np.ones(n_obj)
        A = np.linalg.solve(S, b)
        A[A==0] = 0.000001
        A = 1 / A
        A = A.T
    except LinAlgError:

        A = np.max(S, axis=0)
    A[A == 0] = 0.000001
    F = F / A

    if not return_bounds:
        return F
    else:
        x_max = A + x_min
        return F, x_min, x_max

def standardize(x, return_bounds=False):
    mean = np.mean(x, axis=0)
    std = np.std(x, axis=0)

    # standardize
    val = (x - mean) / std

    if not return_bounds:
        return val
    else:
        return val, mean, std


def destandardize(x, mean, std):
    return (x * std) + mean

# returns only the unique rows from a given matrix X
def unique_rows(X):
    y = np.ascontiguousarray(X).view(np.dtype((np.void, X.dtype.itemsize * X.shape[1])))
    _, idx = np.unique(y, return_index=True)
    return idx


# repairs a numpy array to be in bounds
def repair(X, xl, xu):
    larger_than_xu = X[0, :] > xu
    X[0, larger_than_xu] = xu[larger_than_xu]

    smaller_than_xl = X[0, :] < xl
    X[0, smaller_than_xl] = xl[smaller_than_xl]

    return X


def parameter_less_constraints(F, CV, F_max=None):
    if F_max is None:
        F_max = np.max(F)
    has_constraint_violation = CV > 0
    F[has_constraint_violation] = CV[has_constraint_violation] + F_max
    return F



from pymoo.rand.impl.my_random_generator import MyRandomGenerator
from pymoo.rand.impl.numpy_random_generator import NumpyRandomGenerator
from pymoo.rand.impl.secure_random_generator import SecureRandomGenerator
from pymoo.rand.impl.default_random_generator import DefaultRandomGenerator

class Configuration:
    EPS = 1e-30
    rand = NumpyRandomGenerator()


from pymoo.configuration import Configuration
from pymoo.model.crossover import Crossover
from pymoo.rand import random


class SimulatedBinaryCrossover(Crossover):
    def __init__(self, eta_xover=15, p_xover=0.9):
        super().__init__(2, 2)
        self.p_xover = p_xover
        self.eta_c = eta_xover

    def _do(self, p, parents, children, data=None):

        n = p.n_var

        if random.random() <= self.p_xover:

            for i in range(n):

                if random.random() <= 0.5:

                    if abs(parents[0, i] - parents[1, i]) > Configuration.EPS:

                        if parents[0, i] < parents[1, i]:
                            y1 = parents[0, i]
                            y2 = parents[1, i]
                        else:
                            y1 = parents[1, i]
                            y2 = parents[0, i]

                        yl = p.xl[i]
                        yu = p.xu[i]
                        rand = random.random()
                        beta = 1.0 + (2.0 * (y1 - yl) / (y2 - y1))
                        alpha = 2.0 - pow(beta, -(self.eta_c + 1.0))
                        if rand <= (1.0 / alpha):
                            betaq = pow((rand * alpha), (1.0 / (self.eta_c + 1.0)))

                        else:
                            betaq = pow((1.0 / (2.0 - rand * alpha)), (1.0 / (self.eta_c + 1.0)))

                        c1 = 0.5 * ((y1 + y2) - betaq * (y2 - y1))
                        beta = 1.0 + (2.0 * (yu - y2) / (y2 - y1))
                        alpha = 2.0 - pow(beta, -(self.eta_c + 1.0))
                        if rand <= (1.0 / alpha):
                            betaq = pow((rand * alpha), (1.0 / (self.eta_c + 1.0)))
                        else:
                            betaq = pow((1.0 / (2.0 - rand * alpha)), (1.0 / (self.eta_c + 1.0)))

                        c2 = 0.5 * ((y1 + y2) + betaq * (y2 - y1))

                        if c1 < yl:
                            c1 = yl
                        if c2 < yl:
                            c2 = yl
                        if c1 > yu:
                            c1 = yu
                        if c2 > yu:
                            c2 = yu
                        if random.random() <= 0.5:
                            children[0, i] = c2
                            children[1, i] = c1

                        else:
                            children[0, i] = c1
                            children[1, i] = c2

                    else:
                        children[0, i] = parents[0, i]
                        children[1, i] = parents[1, i]

                else:
                    children[0, i] = parents[0, i]
                    children[1, i] = parents[1, i]

        else:
            children[:, :] = parents

        return children
